@@ -77,7 +77,6 @@ bool   BIOSEG_PREFIX(_over_right)(SEG * a, SEG * b);
 void   BIOSEG_PREFIX(_rt_size)(SEG * a, int32 *sz);
 int32  BIOSEG_PREFIX(_size)(SEG * a);
 
-static SEG *BIOSEG_PREFIX(_inter)(SEG * a, SEG * b);
 static SEG *BIOSEG_PREFIX(_union)(SEG * a, SEG * b);
 
 /*
@@ -213,7 +212,7 @@ int get_int(char **strp, int32 *result) {
     ereport(ERROR,
             (errcode(ERRCODE_SYNTAX_ERROR),
              errmsg("bad bioseg representation"),
-             errdetail("integer %ld at: %s is out of range - must be >= %d",
+             errdetail("integer %ld at: %s is out of range - must be >= %ld",
                        long_result, *strp, MIN_LOWER)));
     return 0;
   }
@@ -294,6 +293,15 @@ SEG *
     return NULL;
   }
 
+  if (lower < MIN_LOWER) {
+    ereport(ERROR,
+            (errcode(ERRCODE_SYNTAX_ERROR),
+             errmsg("bad bioseg representation"),
+             errdetail("lower bound (%ld) must be >= %ld",
+                       (long int) lower, MIN_LOWER)));
+    return 0;
+  }
+
   result->lower = lower;
   result->upper = upper;
 
@@ -401,18 +409,28 @@ float *
   BIOSEG_PREFIX(_gist_penalty)(GISTENTRY *origentry, GISTENTRY *newentry, float *result)
 {
   SEG   *ud;
+  SEG   *origrange;
+  SEG   *newrange;
   int32 tmp1;
   int32 tmp2;
 
-  ud = BIOSEG_PREFIX(_union)((SEG *) DatumGetPointer(origentry->key),
-                    (SEG *) DatumGetPointer(newentry->key));
+  origrange = (SEG *) DatumGetPointer(origentry->key);
+  newrange = (SEG *) DatumGetPointer(newentry->key);
+  ud = BIOSEG_PREFIX(_union)(origrange, newrange);
   BIOSEG_PREFIX(_rt_size)(ud, &tmp1);
-  BIOSEG_PREFIX(_rt_size)((SEG *) DatumGetPointer(origentry->key), &tmp2);
-  *result = tmp1 - tmp2;
+  BIOSEG_PREFIX(_rt_size)(origrange, &tmp2);
+  if (tmp1 == tmp2)
+    {
+      BIOSEG_PREFIX(_rt_size)(newrange, &tmp1);
+      *result = ((tmp2 - tmp1) * 100000.0) / tmp2;
+    }
+  else
+    {
+      *result = tmp1 - tmp2 + 100000.0;
+    }
 
 #ifdef GIST_DEBUG
-  fprintf(stderr, "penalty\n");
-  fprintf(stderr, "\t%g\n", *result);
+  fprintf(stderr, "Penalty of putting %d..%d into %d..%d is %g\n", newrange->lower, newrange->upper, origrange->lower, origrange->upper, *result);
 #endif
 
   return (result);
@@ -421,143 +439,87 @@ float *
 
 /*
 ** The GiST PickSplit method for segments
-** We use Guttman's poly time split algorithm
 */
 GIST_SPLITVEC *
   BIOSEG_PREFIX(_gist_picksplit)(GistEntryVector *entryvec,
                     GIST_SPLITVEC *v)
 {
   OffsetNumber  i;
-  OffsetNumber  j;
   SEG          *datum_alpha;
-  SEG          *datum_beta;
   SEG          *datum_l;
   SEG          *datum_r;
-  SEG          *union_d;
-  SEG          *union_dl;
-  SEG          *union_dr;
-  SEG          *inter_d;
-  bool          firsttime;
-  int32         size_alpha;
-  int32         size_beta;
-  int32         size_union;
-  int32         size_inter;
-  int32         size_waste;
-  int32         waste;
-  int32         size_l;
-  int32         size_r;
+  int32         lowest;
+  int32         highest;
+  int32         midpoint;
+  int32         split;
   int           nbytes;
-  OffsetNumber  seed_1 = 1;
-  OffsetNumber  seed_2 = 2;
+  int           firsttime;
+  long long int midpointsum;
+  int           sumcount;
   OffsetNumber *left;
   OffsetNumber *right;
   OffsetNumber  maxoff;
 
-#ifdef GIST_DEBUG
-  fprintf(stderr, "picksplit\n");
-#endif
-
-  maxoff = entryvec->n - 2;
+  maxoff = entryvec->n - 1;
   nbytes = (maxoff + 2) * sizeof(OffsetNumber);
   v->spl_left = (OffsetNumber *) palloc(nbytes);
   v->spl_right = (OffsetNumber *) palloc(nbytes);
 
+  midpointsum = 0;
+  sumcount = 0;
   firsttime = true;
-  waste = 0.0;
+  lowest = 0;
+  highest = 0;
 
-  for (i = FirstOffsetNumber; i < maxoff; i = OffsetNumberNext(i))
+  for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
     {
       datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[i].key);
-      for (j = OffsetNumberNext(i); j <= maxoff; j = OffsetNumberNext(j))
+      midpoint = datum_alpha->lower + ((datum_alpha->upper - datum_alpha->lower) / 2);
+      midpointsum += midpoint;
+      sumcount++;
+      if (firsttime || (midpoint < lowest))
         {
-          datum_beta = (SEG *) DatumGetPointer(entryvec->vector[j].key);
-
-          /* compute the wasted space by unioning these guys */
-          /* size_waste = size_union - size_inter; */
-          union_d = BIOSEG_PREFIX(_union)(datum_alpha, datum_beta);
-          BIOSEG_PREFIX(_rt_size)(union_d, &size_union);
-          inter_d = BIOSEG_PREFIX(_inter)(datum_alpha, datum_beta);
-          BIOSEG_PREFIX(_rt_size)(inter_d, &size_inter);
-          size_waste = size_union - size_inter;
-
-          /*
-           * are these a more promising split that what we've already seen?
-           */
-          if (size_waste > waste || firsttime)
-            {
-              waste = size_waste;
-              seed_1 = i;
-              seed_2 = j;
-              firsttime = false;
-            }
+          lowest = midpoint;
         }
+      if (firsttime || (midpoint > highest))
+        {
+          highest = midpoint;
+        }
+      firsttime = false;
     }
 
+  split = midpointsum / sumcount;
   left = v->spl_left;
   v->spl_nleft = 0;
   right = v->spl_right;
   v->spl_nright = 0;
 
-  datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[seed_1].key);
-  datum_l = BIOSEG_PREFIX(_union)(datum_alpha, datum_alpha);
-  BIOSEG_PREFIX(_rt_size)(datum_l, &size_l);
-  datum_beta = (SEG *) DatumGetPointer(entryvec->vector[seed_2].key);
-  datum_r = BIOSEG_PREFIX(_union)(datum_beta, datum_beta);
-  BIOSEG_PREFIX(_rt_size)(datum_r, &size_r);
+  datum_l = (SEG *) palloc(sizeof(*datum_l));
+  datum_l->lower = lowest;
+  datum_l->upper = lowest;
+  datum_r = (SEG *) palloc(sizeof(*datum_r));
+  datum_r->lower = highest;
+  datum_r->upper = highest;
 
   /*
-   * Now split up the regions between the two seeds.      An important property
-   * of this split algorithm is that the split vector v has the indices of
-   * items to be split in order in its left and right vectors.  We exploit
-   * this property by doing a merge in the code that actually splits the
-   * page.
-   *
-   * For efficiency, we also place the new index tuple in this loop. This is
-   * handled at the very end, when we have placed all the existing tuples
-   * and i == maxoff + 1.
+   * Now split up the regions between the two seeds.
    */
 
-  maxoff = OffsetNumberNext(maxoff);
   for (i = FirstOffsetNumber; i <= maxoff; i = OffsetNumberNext(i))
     {
-      /*
-       * If we've already decided where to place this item, just put it on
-       * the right list.      Otherwise, we need to figure out which page needs
-       * the least enlargement in order to store the item.
-       */
-
-      if (i == seed_1)
-        {
-          *left++ = i;
-          v->spl_nleft++;
-          continue;
-        }
-      else if (i == seed_2)
-        {
-          *right++ = i;
-          v->spl_nright++;
-          continue;
-        }
-
-      /* okay, which page needs least enlargement? */
       datum_alpha = (SEG *) DatumGetPointer(entryvec->vector[i].key);
-      union_dl = BIOSEG_PREFIX(_union)(datum_l, datum_alpha);
-      union_dr = BIOSEG_PREFIX(_union)(datum_r, datum_alpha);
-      BIOSEG_PREFIX(_rt_size)(union_dl, &size_alpha);
-      BIOSEG_PREFIX(_rt_size)(union_dr, &size_beta);
+      midpoint = datum_alpha->lower + ((datum_alpha->upper - datum_alpha->lower) / 2);
 
       /* pick which page to add it to */
-      if (size_alpha - size_l < size_beta - size_r)
+      if (midpoint <= split)
         {
-          datum_l = union_dl;
-          size_l = size_alpha;
+          datum_l = BIOSEG_PREFIX(_union)(datum_l, datum_alpha);
           *left++ = i;
           v->spl_nleft++;
         }
       else
         {
-          datum_r = union_dr;
-          size_r = size_alpha;
+          datum_r = BIOSEG_PREFIX(_union)(datum_r, datum_alpha);
           *right++ = i;
           v->spl_nright++;
         }
@@ -567,6 +529,11 @@ GIST_SPLITVEC *
   v->spl_ldatum = PointerGetDatum(datum_l);
   v->spl_rdatum = PointerGetDatum(datum_r);
 
+#ifdef GIST_DEBUG
+  fprintf(stderr, "Picksplit summary to %d..%d (%d locations) and %d..%d (%d locations), with pivot %d\n", datum_l->lower, datum_l->upper, v->spl_nleft,
+datum_r->lower, datum_r->upper, v->spl_nright, split);
+#endif
+
   return v;
 }
 
@@ -598,10 +565,6 @@ bool
 {
   bool retval;
 
-#ifdef GIST_QUERY_DEBUG
-  fprintf(stderr, "leaf_consistent, %d\n", strategy);
-#endif
-
   switch (strategy)
     {
     case RTLeftStrategyNumber:
@@ -633,6 +596,18 @@ bool
     default:
       retval = FALSE;
     }
+
+#ifdef GIST_QUERY_DEBUG
+  if (retval)
+    {
+      fprintf(stderr, "leaf_consistent, %d..%d %d %d..%d matches\n", query->lower, query->upper, strategy, key->lower, key->upper);
+    }
+  else
+    {
+      fprintf(stderr, "leaf_consistent, %d..%d %d %d..%d\n", query->lower, query->upper, strategy, key->lower, key->upper);
+    }
+#endif
+
   return (retval);
 }
 
@@ -643,10 +618,6 @@ bool
 {
   bool retval;
 
-#ifdef GIST_QUERY_DEBUG
-  fprintf(stderr, "internal_consistent, %d\n", strategy);
-#endif
-
   switch (strategy)
     {
     case RTLeftStrategyNumber:
@@ -676,6 +647,18 @@ bool
     default:
       retval = FALSE;
     }
+
+#ifdef GIST_QUERY_DEBUG
+  if (retval)
+    {
+      fprintf(stderr, "internal_consistent, %d..%d %d %d..%d matches\n", query->lower, query->upper, strategy, key->lower, key->upper);
+    }
+  else
+    {
+      fprintf(stderr, "internal_consistent, %d..%d %d %d..%d\n", query->lower, query->upper, strategy, key->lower, key->upper);
+    }
+#endif
+
   return (retval);
 }
 
@@ -802,37 +785,6 @@ static SEG *
   return (n);
 }
 
-
-static SEG *
-  BIOSEG_PREFIX(_inter)(SEG * a, SEG * b)
-{
-  SEG *n;
-
-  n = (SEG *) palloc(sizeof(*n));
-
-  /* take min of upper endpoints */
-  if (a->upper < b->upper)
-    {
-      n->upper = a->upper;
-    }
-  else
-    {
-      n->upper = b->upper;
-    }
-
-  /* take max of lower endpoints */
-  if (a->lower > b->lower)
-    {
-      n->lower = a->lower;
-    }
-  else
-    {
-      n->lower = b->lower;
-    }
-
-  return (n);
-}
-
 void
   BIOSEG_PREFIX(_rt_size)(SEG * a, int32 *size)
 {
